/*
 * Copyright (C) 2004-2015 Free Software Foundation, Inc.
 * Copyright (C) 2015-2022 Red Hat, Inc.
 *
 * Author: Nikos Mavrogiannopoulos
 *
 * This file is part of GnuTLS.
 *
 * The GnuTLS is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>
 *
 */

/* Code split off from priority.c, `struct cfg` and some operations on it */

#ifndef GNUTLS_LIB_CONFIG_INT_H
#define GNUTLS_LIB_CONFIG_INT_H

/*
 * struct cfg
 */

struct cfg {
	bool allowlisting;

	name_val_array_t priority_strings;
	char *priority_string;
	char *default_priority_string;
	gnutls_certificate_verification_profiles_t verification_profile;

	gnutls_cipher_algorithm_t ciphers[MAX_ALGOS+1];
	gnutls_mac_algorithm_t macs[MAX_ALGOS+1];
	gnutls_group_t groups[MAX_ALGOS+1];
	gnutls_kx_algorithm_t kxs[MAX_ALGOS+1];
	gnutls_sign_algorithm_t sigs[MAX_ALGOS+1];
	gnutls_protocol_t versions[MAX_ALGOS+1];

	gnutls_digest_algorithm_t hashes[MAX_ALGOS+1];
	gnutls_ecc_curve_t ecc_curves[MAX_ALGOS+1];
	gnutls_sign_algorithm_t sigs_for_cert[MAX_ALGOS+1];
};

/*
 * deinit / partial duplication. no initialization, must be zero-initialized
 */

static inline void
cfg_deinit(struct cfg *cfg)
{
	if (cfg->priority_strings) {
		_name_val_array_clear(&cfg->priority_strings);
	}
	gnutls_free(cfg->priority_string);
	gnutls_free(cfg->default_priority_string);
}

static inline void
cfg_steal(struct cfg *dst, struct cfg *src)
{
	dst->verification_profile = src->verification_profile;

	dst->priority_strings = src->priority_strings;
	src->priority_strings = NULL;

	dst->priority_string = src->priority_string;
	src->priority_string = NULL;

	dst->default_priority_string = src->default_priority_string;
	src->default_priority_string = NULL;

	dst->allowlisting = src->allowlisting;
	memcpy(dst->ciphers, src->ciphers, sizeof(src->ciphers));
	memcpy(dst->macs, src->macs, sizeof(src->macs));
	memcpy(dst->groups, src->groups, sizeof(src->groups));
	memcpy(dst->kxs, src->kxs, sizeof(src->kxs));
	memcpy(dst->hashes, src->hashes, sizeof(src->hashes));
	memcpy(dst->ecc_curves, src->ecc_curves, sizeof(src->ecc_curves));
	memcpy(dst->sigs, src->sigs, sizeof(src->sigs));
	memcpy(dst->sigs_for_cert, src->sigs_for_cert,
	       sizeof(src->sigs_for_cert));
}

/*
 * synchronizing changes from struct cfg to global `lib/algorithms` arrays
 */

/* global side-effect! modifies `flags` in `hash_algorithms[]` */
static inline int /* allowlisting-only */
_cfg_hashes_remark(struct cfg* cfg)
{
	size_t i;
	_gnutls_digest_mark_insecure_all();
	for (i = 0; cfg->hashes[i] != 0; i++) {
		int ret = _gnutls_digest_set_secure(cfg->hashes[i], 1);
		if (unlikely(ret < 0)) {
			return gnutls_assert_val(ret);
		}
	}
	return 0;
}

/* global side-effect! modifies `flags` in `sign_algorithms[]` */
static inline int /* allowlisting-only */
_cfg_sigs_remark(struct cfg* cfg)
{
	size_t i;
	_gnutls_sign_mark_insecure_all(_INSECURE);
	for (i = 0; cfg->sigs[i] != 0; i++) {
		int ret = _gnutls_sign_set_secure(cfg->sigs[i],
						  _INSECURE_FOR_CERTS);
		if (unlikely(ret < 0)) {
			return gnutls_assert_val(ret);
		}
	}
	for (i = 0; cfg->sigs_for_cert[i] != 0; i++) {
		int ret = _gnutls_sign_set_secure(cfg->sigs_for_cert[i],
		                                  _SECURE);
		if (unlikely(ret < 0)) {
			return gnutls_assert_val(ret);
		}
	}
	return 0;
}

/* global side-effect! modifies `supported` in `sup_versions[]` */
static inline int /* allowlisting-only */
_cfg_versions_remark(struct cfg* cfg)
{
	size_t i;
	_gnutls_version_mark_disabled_all();
	for (i = 0; cfg->versions[i] != 0; i++) {
		int ret = _gnutls_protocol_set_enabled(cfg->versions[i], 1);
		if (unlikely(ret < 0)) {
			return gnutls_assert_val(ret);
		}
	}
	return 0;
}

/* global side-effect! modifies `supported` in `ecc_curves[]` */
static inline int /* allowlisting-only */
_cfg_ecc_curves_remark(struct cfg* cfg)
{
	size_t i;
	_gnutls_ecc_curve_mark_disabled_all();
	for (i = 0; cfg->ecc_curves[i] != 0; i++) {
		int ret = _gnutls_ecc_curve_set_enabled(cfg->ecc_curves[i], 1);
		if (unlikely(ret < 0)) {
			return gnutls_assert_val(ret);
		}
	}
	return 0;
}

/*
 * setting arrays of struct cfg: from other arrays
 */

static inline int /* allowlisting-only */
cfg_hashes_set_array(struct cfg* cfg,
                     gnutls_digest_algorithm_t* src, size_t len)
{
	if (unlikely(len >= MAX_ALGOS)) {
		return gnutls_assert_val(GNUTLS_A_INTERNAL_ERROR);
	}
	if (len) {
		memcpy(cfg->hashes,
		       src, sizeof(gnutls_digest_algorithm_t) * len);
	}
	cfg->hashes[len] = 0;
	return _cfg_hashes_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_sigs_set_arrays(struct cfg* cfg,
		    gnutls_sign_algorithm_t* src, size_t len,
		    gnutls_sign_algorithm_t* src_for_cert, size_t len_for_cert)
{
	if (unlikely(len >= MAX_ALGOS)) {
		return gnutls_assert_val(GNUTLS_A_INTERNAL_ERROR);
	}
	if (unlikely(len_for_cert >= MAX_ALGOS)) {
		return gnutls_assert_val(GNUTLS_A_INTERNAL_ERROR);
	}
	if (len) {
		memcpy(cfg->sigs, src, sizeof(gnutls_sign_algorithm_t) * len);
	}
	if (len_for_cert) {
		memcpy(cfg->sigs_for_cert, src_for_cert,
		       sizeof(gnutls_sign_algorithm_t) * len_for_cert);
	}
	cfg->sigs[len] = 0;
	cfg->sigs_for_cert[len_for_cert] = 0;
	return _cfg_sigs_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_versions_set_array(struct cfg* cfg, gnutls_protocol_t* src, size_t len)
{
	if (unlikely(len >= MAX_ALGOS)) {
		return gnutls_assert_val(GNUTLS_A_INTERNAL_ERROR);
	}
	if (len) {
		memcpy(cfg->versions, src, sizeof(gnutls_protocol_t) * len);
	}
	cfg->versions[len] = 0;
	return _cfg_versions_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_ecc_curves_set_array(struct cfg* cfg, gnutls_ecc_curve_t* src, size_t len)
{
	if (unlikely(len >= MAX_ALGOS)) {
		return gnutls_assert_val(GNUTLS_A_INTERNAL_ERROR);
	}
	if (len) {
		memcpy(cfg->ecc_curves, src, sizeof(gnutls_ecc_curve_t) * len);
	}
	cfg->ecc_curves[len] = 0;
	return _cfg_ecc_curves_remark(cfg);
}

/*
 * appending to arrays of struct cfg
 */

/* polymorphic way to DRY this operation. other possible approaches:
 * 1. just unmacro (long)
 * 2. cast to ints and write a function operating on ints
 *    (hacky, every call is +4 lines, needs a portable static assert)
 * 3. macro whole functions, not just this operation (harder to find/read)
 */
#define APPEND_TO_NULL_TERMINATED_ARRAY(dst, element) \
	do { \
		size_t i; \
		for (i = 0; dst[i] != 0; i++) { \
			if (dst[i] == element) { \
				return 0; \
			} \
		} \
		if (unlikely(i >= MAX_ALGOS)) { \
			return gnutls_assert_val(GNUTLS_A_INTERNAL_ERROR); \
		} \
		dst[i] = element; \
		dst[i + 1] = 0; \
	} while (0)

static inline int /* allowlisting-only */
cfg_hashes_add(struct cfg *cfg, gnutls_digest_algorithm_t dig)
{
	_gnutls_debug_log("cfg: enabling digest algorithm %s\n",
	                  gnutls_digest_get_name(dig));
	APPEND_TO_NULL_TERMINATED_ARRAY(cfg->hashes, dig);
	return _cfg_hashes_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_sigs_add(struct cfg *cfg, gnutls_sign_algorithm_t sig)
{
	_gnutls_debug_log("cfg: enabling signature algorithm "
	                  "(for non-certificate usage) "
	                  "%s\n", gnutls_sign_get_name(sig));
	APPEND_TO_NULL_TERMINATED_ARRAY(cfg->sigs, sig);
	return _cfg_sigs_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_sigs_for_cert_add(struct cfg *cfg, gnutls_sign_algorithm_t sig)
{
	_gnutls_debug_log("cfg: enabling signature algorithm"
	                  "(for certificate usage) "
	                  "%s\n", gnutls_sign_get_name(sig));
	APPEND_TO_NULL_TERMINATED_ARRAY(cfg->sigs_for_cert, sig);
	return _cfg_sigs_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_versions_add(struct cfg *cfg, gnutls_protocol_t prot)
{
	_gnutls_debug_log("cfg: enabling version %s\n",
	                  gnutls_protocol_get_name(prot));
	APPEND_TO_NULL_TERMINATED_ARRAY(cfg->versions, prot);
	return _cfg_versions_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_ecc_curves_add(struct cfg *cfg, gnutls_ecc_curve_t curve)
{
	_gnutls_debug_log("cfg: enabling curve %s\n",
	                  gnutls_ecc_curve_get_name(curve));
	APPEND_TO_NULL_TERMINATED_ARRAY(cfg->ecc_curves, curve);
	return _cfg_ecc_curves_remark(cfg);
}

#undef APPEND_TO_NULL_TERMINATED_ARRAY

/*
 * removing from arrays of struct cfg
 */

/* polymorphic way to DRY this removal, see APPEND_TO_NULL_TERMINATED_ARRAY */
#define REMOVE_FROM_NULL_TERMINATED_ARRAY(dst, element) \
	do { \
		size_t i, j; \
		for (i = 0; dst[i] != 0; i++) { \
			if (dst[i] == element) { \
				for (j = i; dst[j] != 0; j++) { \
					dst[j] = dst[j + 1]; \
				} \
			} \
		} \
	} while (0)

static inline int /* allowlisting-only */
cfg_hashes_remove(struct cfg *cfg, gnutls_digest_algorithm_t dig)
{
	_gnutls_debug_log("cfg: disabling digest algorithm %s\n",
	                  gnutls_digest_get_name(dig));
	REMOVE_FROM_NULL_TERMINATED_ARRAY(cfg->hashes, dig);
	return _cfg_hashes_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_sigs_remove(struct cfg *cfg, gnutls_sign_algorithm_t sig)
{
	_gnutls_debug_log("cfg: disabling signature algorithm "
	                  "(for non-certificate usage) "
	                  "%s\n", gnutls_sign_get_name(sig));
	REMOVE_FROM_NULL_TERMINATED_ARRAY(cfg->sigs, sig);
	return _cfg_sigs_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_sigs_for_cert_remove(struct cfg *cfg, gnutls_sign_algorithm_t sig)
{
	_gnutls_debug_log("cfg: disabling signature algorithm"
	                  "(for certificate usage) "
	                  "%s\n", gnutls_sign_get_name(sig));
	REMOVE_FROM_NULL_TERMINATED_ARRAY(cfg->sigs_for_cert, sig);
	return _cfg_sigs_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_versions_remove(struct cfg *cfg, gnutls_protocol_t prot)
{
	_gnutls_debug_log("cfg: disabling version %s\n",
	                  gnutls_protocol_get_name(prot));
	REMOVE_FROM_NULL_TERMINATED_ARRAY(cfg->versions, prot);
	return _cfg_versions_remark(cfg);
}

static inline int /* allowlisting-only */
cfg_ecc_curves_remove(struct cfg *cfg, gnutls_ecc_curve_t curve)
{
	_gnutls_debug_log("cfg: disabling curve %s\n",
	                  gnutls_ecc_curve_get_name(curve));
	REMOVE_FROM_NULL_TERMINATED_ARRAY(cfg->ecc_curves, curve);
	return _cfg_ecc_curves_remark(cfg);
}

#endif  /* GNUTLS_LIB_CONFIG_INT_H */
